// -*- C++ -*-
/*!
 * @file  CorbaPort.h
 * @brief CorbaPort class
 * @date  $Date: 2007/01/04 00:45:21 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: CorbaPort.h,v 1.1 2007/01/04 00:45:21 n-ando Exp $
 *
 */

/*
 * $Log: CorbaPort.h,v $
 * Revision 1.1  2007/01/04 00:45:21  n-ando
 * CORBA serivce provider/consumer Port implementation.
 *
 *
 */


#ifndef CorbaPort_h
#define CorbaPort_h

#include <rtm/PortBase.h>
#include <rtm/CorbaConsumer.h>
#include <rtm/NVUtil.h>

namespace RTC
{
  /*!
   * @if jp
   * @class CorbaPort
   * @brief RT ݡͥ CORBA service/consumer  Port
   *
   * CorbaPort  RT ݡͥȤˤơ桼 CORBA ֥
   * ӥӥ󥷥塼ޤ󶡤 Port Ǥ롣
   * <p>
   * RT ݡͥȤϡPort 𤷤ƥ桼 CORBA ӥ
   * 뤳ȤǤ RT Service (Provider) ȸƤ֡
   * ޤ¾ RT ݡͥȤΥӥѤ뤿 CORBA ֥
   * Υץ졼ۥ󶡤뤳ȤǤ RT Service Consumer ȸƤ֡
   * <p>
   * CorbaPort Ǥդο Provider  Consumer 뤳ȤǤ
   * Port ƱΤ³ݤб Provider  Consumer Ŭڤ˴Ϣդ
   * ȤǤ롣
   * <p>
   * CorbaPort ̾ʲΤ褦Ѥ롣
   *
   * <pre>
   * RTC::CorbaPort m_port0; // Port 
   *
   * MyService_impl m_mysvc0; //  Port 󶡤 Serivce Provider
   * RTC::CorbaConsumer<YourService> m_cons0; //  Port  Consumer
   *
   * // Service Provider  Port Ͽ
   * m_port0.registerProvider("MyService0", "Generic", m_mysvc0);
   * // Service Consumer  Port Ͽ
   * m_port0.registerConsumer("YourService0", "Generic", m_cons0 );
   *
   * // connect Ԥ줿
   *
   * m_cons0->your_service_function(); // YourService δؿ򥳡
   *
   * // connect 줿 ̤ΥݡͥȤˤ
   * m_cons1->my_service_function(); // MyService δؿ򥳡
   * </pre>
   *
   * Τ褦ˡ󶡤 Service Provider  registerProvider() Ͽ
   * 뤳Ȥˤꡢ¾ΥݡͥȤѲǽˤ¾
   * Ѥ Service Consumer  registerConsumer() Ͽ뤳Ȥˤ
   * ¾ΥݡͥȤ Service 򥳥ݡͥѲǽˤ뤳Ȥ
   * Ǥ롣
   *
   * @else
   * @class CorbaPort
   * @brief RT Conponent CORBA service/consumer Port
   *
   * CorbaPort is an implementation of the Port of RT-Component's that provides
   * user-defined CORBA Object Service and Consumer.
   * <p>
   * RT-Component can provide user-defined CORBA serivces, which is called
   * RT-Serivce (Provider), through the Ports.
   * RT-Component can also provide place-holder, which is called RT-Serivce
   * Consumer, to use other RT-Component's service.
   * <p>
   * The CorbaPort can manage any number of Providers and Consumers, can
   * associate Consumers with correspondent Providers when establishing
   * connection among Ports.
   * <p>
   * Usually, CorbaPort is used like the following.
   *
   * <pre>
   * RTC::CorbaPort m_port0; // declaration of Port
   *
   * MyService_impl m_mysvc0; // Serivce Provider that is provided by the Port
   * RTC::CorbaConsumer<YourService> m_cons0; // Consumer of the Port
   *
   * // register Service Provider to the Port
   * m_port0.registerProvider("MyService0", "Generic", m_mysvc0);
   * // register Service Consumer to the Port
   * m_port0.registerConsumer("YourService0", "Generic", m_cons0 );
   *
   * // after connect established
   *
   * m_cons0->your_service_function(); // call a YourService's function
   *
   * // in another component that is connected with the Port
   * m_cons1->my_service_function(); // call a MyService's function
   * </pre>
   *
   * Registering Service Provider by registerProvider(), it can be used from
   * other RT-Components.
   * Registering Service Consumer by registerConsumer(), other RT-Component's
   * services can be used through the consumer object.
   *
   * @endif
   */
  class CorbaPort
    : public PortBase
  {
  public:
    /*!
     * @if jp
     * @brief 󥹥ȥ饯
     *
     * @param name Port ̾
     *
     * @else
     *
     * @brief Constructor
     *
     * @param name The name of Port 
     *
     * @endif
     */
    CorbaPort(const char* name);


    /*!
     * @if jp
     *
     * @brief ǥȥ饯
     *
     * @else
     *
     * @brief Destructor
     *
     * @endif
     */
    virtual ~CorbaPort();


    /*!
     * @if jp
     *
     * @brief Provider Ͽ
     *
     *  Port ˤ󶡤ХȤ򤳤 Port ФϿ롣
     * ХȤϡͿ instance_name, type_name 
     * ХȼȤΥ󥹥̾ӥ̾ȤơХȤ
     * Ϣդ롣
     *
     * @param instance_name ХȤΥ󥹥̾
     * @param type_name ХȤΥ̾
     * @param provider CORBA Х
     * @return Ʊ̾ instance_name ϿƤ false ֤
     *
     * @else
     *
     * @brief Register provider
     *
     * This operation registers a servant, which is provided in this Port,
     * to the Port. The servant is associated with "instance_name" and
     * "type_name" as the instance name of the servant and as the type name
     * of the servant.
     *
     * @param instance_name ХȤΥ󥹥̾
     * @param type_name ХȤΥ̾
     * @param provider CORBA Х
     * @return Ʊ̾ instance_name ϿƤ false ֤
     *
     * @endif
     */
    bool registerProvider(const char* instance_name, const char* type_name,
			  PortableServer::RefCountServantBase& provider);


    /*!
     * @if jp
     *
     * @brief Consumer Ͽ
     *
     *  Port ׵᤹륵ӥΥץ졼ۥǤ륳󥷥塼
     * (Consumer) Ͽ롣
     * Consumer Ϣդ륵ӥΥ󥹥̾ӥ̾Ȥơ
     *  instance_name, type_name  Consumer ȤͿ뤳Ȥˤꡢ
     * Ǥ餬Ϣդ롣
     * Port ֤³ (connect)  ˤϡƱ instance_name, type_name 
     * ӥ¾ Port  (Provide) Ƥ硢Υӥ
     * ֥ȻȤưŪ Consumer ˥åȤ롣
     *
     * @param instance_name Consumer ׵᤹륵ӥΥ󥹥̾
     * @param type_name Consumer ׵᤹륵ӥΥ̾
     * @param consumer CORBA ӥ󥷥塼
     * @return Ʊ̾ instance_name ϿƤ false ֤
     *
     * @else
     *
     * @brief Register consumer
     *
     * This operation registers a consumer, which requiers a service,
     * to the other Port. The consumer is associated with "instance_name" and
     * "type_name" as the instance name of the service and as the type name
     * of the service that is required.
     *
     * @param instance_name An instance name of the service required
     * @param type_name An type name of the service required
     * @param consumer CORBA service consumer
     * @return False would be returned if the same instance_name is registered
     *
     * @endif
     */
    bool registerConsumer(const char* instance_name, const char* type_name,
			  CorbaConsumerBase& consumer);


  protected:
    /*!
     * @if jp
     *
     * @brief Interface 
     *
     *  Portͭ Provider ˴ؤ ConnectorProfile::properties
     * 롣
     * ϡNVList name  value ȤưʲΤΤǼ롣
     *
     * - port.<type_name>.<instance_name>: <CORBA::Object_ptr>
     *
     * ǡ
     * - <type_name>: PortInterfaceProfile::type_name
     * - <instance_name>: PortInterfaceProfile::instance_name
     * Ǥ롣
     * ConnectorProfile::properties Ǥϡ .(ɥå)ɽǡ
     * NameValue ΥȤƤ롣äơ
     *
     * <pre>
     *  PortInterfaceProfile
     *  {
     *    instance_name = "PA10_0";
     *    type_name     = "Manipulator";
     *    polarity      = PROVIDED;
     *  }
     *</pre>
     *
     * ʤС
     *
     * <pre>
     * NameValue = { "port.Manipulator.PA10_0": <Object reference> }
     * </pre>
     *
     * Ȥäͤ ConnectorProfile::properties ˳Ǽ졢¾ΥݡȤФ
     * ã롣¾ Port ǤΥ󥿡եѤ Consumer 
     * ¸ߤСConnectorProfile 餳Υ饪֥ȥե󥹤
     * 餫ηǻѤ롣
     *
     * @else
     *
     * @brief Publish interface information
     *
     * @endif
     */
     virtual ReturnCode_t
     publishInterfaces(ConnectorProfile& connector_profile);


    /*!
     * @if jp
     *
     * @brief Interface ³
     *
     *  Portͭ Consumer Ŭ礹 Provider ˴ؤ
     * ConnectorProfile::properties Ф Consumer ˥֥Ȼ
     * 򥻥åȤ롣
     *
     * Consumer 
     * <pre>
     *  PortInterfaceProfile
     *  {
     *    instance_name = "PA10_0";
     *    type_name     = "Manipulator";
     *    polarity      = REQUIRED;
     *  }
     *</pre>
     * ȤϿƤС¾ Port 
     * <pre>
     *  PortInterfaceProfile
     *  {
     *    instance_name = "PA10_0";
     *    type_name     = "Manipulator";
     *    polarity      = PROVIDED;
     *  }
     * </pre> 
     * ȤϿƤ Serivce Provider Υ֥ȻȤõ
     * Consumer ˥åȤ롣
     * ºݤˤϡConnectorProfile::properties 
     * <pre>
     * NameValue = { "port.Manipulator.PA10_0": <Object reference> }
     * </pre>
     * ȤϿƤ NameValue õΥ֥ȻȤ
     * Consumer ˥åȤ롣
     *
     * @else
     *
     * @brief Subscribe interfaces
     *
     * @endif
     */
     virtual ReturnCode_t
     subscribeInterfaces(const ConnectorProfile& connector_profile);


    /*!
     * @if jp
     *
     * @brief Interface ؤ³
     *
     * Ϳ줿 ConnectorProfile ˴Ϣ Consumer ˥åȤ줿
     * ٤Ƥ Object ³롣
     *
     * @else
     *
     * @brief Unsubscribe interfaces
     *
     * @endif
     */
     virtual void
     unsubscribeInterfaces(const ConnectorProfile& connector_profile);


  private:
    /*!
     * @if jp
     * @brief Provider ξǼ NVList
     * @else
     * @brief NVList to be stored Providers' information
     * @endif
     */
    NVList m_providers;


    /*!
     * @if jp
     * @brief Consumer ξǼ빽¤
     * @else
     * @brief Consumer inforamtion struct
     * @endif
     */
    struct Consumer
    {
      Consumer(const char* _instance_name, const char* _type_name,
	       CorbaConsumerBase& _cons)
	: name(std::string("port.")
	       + _type_name
	       + std::string(".")
	       + _instance_name),
	  consumer(_cons)
      {}
      Consumer(const Consumer& cons)
	: name(cons.name), consumer(cons.consumer)
      {	
      }
      Consumer operator=(const Consumer& _cons)
      {
	Consumer cons(_cons);
	return cons;
      }
      std::string name;
      CorbaConsumerBase& consumer;
    };
    std::vector<Consumer> m_consumers;

    // functors
    /*!
     * @if jp
     * @brief ConnectorProfile  Consuemr Ӥ򤷥֥ȻȤ
     *        åȤ뤿 Functor
     * @else
     * @brief Subscription mutching functor for Consumer
     * @endif
     */
    struct subscribe
    {
      subscribe(std::vector<Consumer>& cons)
	: m_cons(cons), m_len(cons.size())
	{
	}
	
	void operator()(const SDOPackage::NameValue& nv)
	{
	  for (CORBA::ULong i = 0; i < m_len; ++i)
	    {
	      std::string name(nv.name);
	      if (m_cons[i].name == name)
		{
		  CORBA::Object_ptr obj;
		  if (nv.value >>= CORBA::Any::to_object(obj))
		    {
		      m_cons[i].consumer.setObject(obj);
		    }
		}
	}
	}
	std::vector<Consumer> m_cons;
	CORBA::ULong m_len;
    };


    /*!
     * @if jp
     * @brief Consumer Υ֥Ȥ뤿 Functor
     * @else
     * @brief Unsubscription functor for Consumer
     * @endif
     */
    struct unsubscribe
    {
      unsubscribe(std::vector<Consumer>& cons)
	: m_cons(cons), m_len(cons.size())
      {
      }
      
      void operator()(const SDOPackage::NameValue& nv)
      {
	for (CORBA::ULong i = 0; i < m_len; ++i)
	  {
	    std::string name(nv.name);
	    if (m_cons[i].name == name)
	      {
		m_cons[i].consumer.releaseObject();
	      }
	  }
      }
      std::vector<Consumer> m_cons;
      CORBA::ULong m_len;
    };
    
  };
};
#endif // CorbaPort_h
