// -*- C++ -*-
/*!
 * @file DataOutPort.h
 * @brief Base class of OutPort
 * @date $Date: 2007/04/13 15:45:08 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: DataOutPort.h,v 1.7 2007/04/13 15:45:08 n-ando Exp $
 *
 */

/*
 * $Log: DataOutPort.h,v $
 * Revision 1.7  2007/04/13 15:45:08  n-ando
 * RTC::OK was changed to RTC::RTC_OK.
 *
 * Revision 1.6  2007/01/21 09:45:31  n-ando
 * To advertise interface profile, publishInterfaceProfile() is called now.
 *
 * Revision 1.5  2007/01/14 22:57:54  n-ando
 * A bug fix about template argument for buffer-type in constructor.
 *
 * Revision 1.4  2007/01/12 14:30:01  n-ando
 * A trivial bug fix.
 *
 * Revision 1.3  2007/01/06 17:44:11  n-ando
 * The behavior on notify_connect() and notify_disconnect() are now
 * implemented in protected functions(ex. publisherInterfaces()).
 *
 * Revision 1.2  2006/12/02 18:29:15  n-ando
 * Now OutPortCorbaProvider and InPortCorbaConsumer are used.
 *
 * Revision 1.1  2006/11/27 09:44:37  n-ando
 * The first commitment.
 *
 *
 */

#ifndef DataOutPort_h
#define DataOutPort_h


#include <rtm/PortBase.h>
#include <rtm/BufferBase.h>
#include <rtm/OutPortCorbaProvider.h>
#include <rtm/OutPortTcpSockProvider.h>
#include <rtm/InPortCorbaConsumer.h>
#include <rtm/InPortTcpSockConsumer.h>
#include <rtm/OutPort.h>
#include <rtm/NVUtil.h>
#include <rtm/PublisherFactory.h>
#include <rtm/Properties.h>

namespace RTC
{
  /*!
   * @if jp
   * @class DataOutPort
   * @brief InPort  Port
   * @else
   * @class DataOutPort
   * @brief InPort abstruct class
   * @endif
   */
  class DataOutPort
    : public PortBase
  {
  public:
    /*!
     * @if jp
     * @brief 󥹥ȥ饯
     * @else
     * @brief Constructor
     * @endif
     */
    template <class DataType, template <class DataType> class Buffer>
    DataOutPort(const char* name, OutPort<DataType, Buffer>& outport,
		Properties& prop)
      : PortBase(name), m_outport(outport)
    {
      // PortProfile::properties 
      addProperty("port.port_type", "DataOutPort");
      
      m_providers.push_back(new OutPortCorbaProvider<DataType>(outport));
      m_providers.back()->publishInterfaceProfile(m_profile.properties);

      m_providers.push_back(new OutPortTcpSockProvider<DataType>(outport));
      m_providers.back()->publishInterfaceProfile(m_profile.properties);

      m_consumers.push_back(new InPortCorbaConsumer<DataType>(outport));
      m_consumers.push_back(new InPortTcpSockConsumer<DataType>(outport, prop));

    }


    /*!
     * @if jp
     * @brief ǥȥ饯
     * @else
     * @brief Destructor
     * @endif
     */
    virtual ~DataOutPort();

    
    /*!
     * @if jp
     * @brief [CORBA interface] Port ³Ԥ
     *
     * OutPort  InPort Ȥ³Ԥ
     *
     * OutPort ¦ connect() ǤϰʲΥ󥹤ǽԤ롣
     *
     * 1. OutPort ˴Ϣ connector ӥå
     *
     * 2. InPort˴Ϣ connector μ
     *  - ConnectorProfile::properties["dataport.corba_any.inport_ref"]
     *    OutPortAny Υ֥ȥե󥹤ꤵƤ硢
     *    ե󥹤Consumer֥Ȥ˥åȤ롣
     *    ե󥹤åȤƤʤ̵뤷Ʒ³
     *    (OutPortconnect() ƤӽФΥȥݥȤξϡ
     *    InPortΥ֥ȥե󥹤ϥåȤƤʤϤǤ롣)
     * 3. PortBase::connect() 򥳡
     *    Port³δܽԤ롣
     * 4. 嵭2.InPortΥե󥹤ǤʤСInPort
     *    Ϣ connector 롣
     *
     * 5. ConnectorProfile::properties Ϳ줿󤫤顢
     *    OutPort¦νԤ
     *
     * - [dataport.interface_type]
     * -- CORBA_Any ξ: 
     *    InPortAny ̤ƥǡ򴹤롣
     *    ConnectorProfile::properties["dataport.corba_any.inport_ref"]
     *    InPortAny Υ֥ȥե󥹤򥻥åȤ롣
     * -- RawTCP ξ: Raw TCP socket ̤ƥǡ򴹤롣
     *    ConnectorProfile::properties["dataport.raw_tcp.server_addr"]
     *    InPort¦ΥХɥ쥹򥻥åȤ롣
     *
     * - [dataport.dataflow_type]
     * -- Pushξ: Subscriber롣SubscriberΥפϡ
     *    dataport.subscription_type ꤵƤ롣
     * -- Pullξ: InPort¦ǡPullǼ뤿ᡢ
     *    ä˲⤹ɬפ̵
     *
     * - [dataport.subscription_type]
     * -- Onceξ: SubscriberOnce롣
     * -- Newξ: SubscriberNew롣
     * -- Periodicξ: SubscriberPeriodic롣
     *
     * - [dataport.push_interval]
     * -- dataport.subscription_type=Periodicξꤹ롣
     *
     * 6. 嵭νΤĤǤ⥨顼ǤС顼꥿󤹤롣
     *    ˽Ԥ줿RTC::RTC_OKǥ꥿󤹤롣
     *  
     * @else
     * @brief [CORBA interface] Connect the Port
     * @endif
     */
    
  protected:
    /*!
     * @if jp
     *
     * @brief Interface 
     *
     * Υڥ졼ϡnotify_connect() 󥹤λϤ˥
     * 貾۴ؿǤ롣
     * notify_connect() Ǥϡ
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * ν protected ؿ뤵³Ԥ롣
     * <br>
     * Υڥ졼ϡ connector_id ФƤ³
     * ¸ connector_id ФƤϹŬڤ˹Ԥɬפ롣
     *
     * @param connector_profile ³˴ؤץե
     * @return ReturnCode_t Υ꥿󥳡
     *
     * @else
     *
     * @brief Publish interface information
     *
     * This operation is pure virutal method that would be called at the
     * beginning of the notify_connect() process sequence.
     * In the notify_connect(), the following methods would be called in order.
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile() 
     *
     * This operation should create the new connection for the new
     * connector_id, and should update the connection for the existing
     * connection_id.
     *
     * @param connector_profile The connection profile information
     * @return The return code of ReturnCode_t type.
     *
     * @endif
     */
    virtual ReturnCode_t
    publishInterfaces(ConnectorProfile& connector_profile);
    

    /*! @if jp
     *
     * @brief Interface 
     *
     * Υڥ졼ϡnotify_connect() 󥹤֤˥
     * 貾۴ؿǤ롣
     * notify_connect() Ǥϡ
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * ν protected ؿ뤵³Ԥ롣
     *
     * @param connector_profile ³˴ؤץե
     * @return ReturnCode_t Υ꥿󥳡
     *
     * @else
     *
     * @brief Publish interface information
     *
     * This operation is pure virutal method that would be called at the
     * mid-flow of the notify_connect() process sequence.
     * In the notify_connect(), the following methods would be called in order.
     *
     * - publishInterfaces()
     * - connectNext()
     * - subscribeInterfaces()
     * - updateConnectorProfile()
     *
     * @param connector_profile The connection profile information
     * @return The return code of ReturnCode_t type.
     *
     * @endif
     */
    virtual ReturnCode_t
    subscribeInterfaces(const ConnectorProfile& connector_profile);
    

    /*!
     * @if jp
     *
     * @brief Interface ³
     *
     * Υڥ졼ϡnotify_disconnect() 󥹤ν˥
     * 貾۴ؿǤ롣
     * notify_disconnect() Ǥϡ
     * - disconnectNext()
     * - unsubscribeInterfaces()
     * - eraseConnectorProfile()
     * ν protected ؿ뤵³Ԥ롣
     *
     * @param connector_profile ³˴ؤץե
     *
     * @else
     *
     * @brief Disconnect interface connection
     *
     * This operation is pure virutal method that would be called at the
     * end of the notify_disconnect() process sequence.
     * In the notify_disconnect(), the following methods would be called.
     * - disconnectNext()
     * - unsubscribeInterfaces()
     * - eraseConnectorProfile() 
     *
     * @param connector_profile The connection profile information
     *
     * @endif
     */
    virtual void
    unsubscribeInterfaces(const ConnectorProfile& connector_profile);


    
    struct publish
    {
      publish(SDOPackage::NVList& prop) : m_prop(prop) {}
      void operator()(OutPortProvider* provider)
      {
	provider->publishInterface(m_prop);
      }
      SDOPackage::NVList& m_prop;
    };

    struct unsubscribe
    {
      unsubscribe(const SDOPackage::NVList& prop) : m_prop(prop) {}
      void operator()(InPortConsumer* consumer)
      {
	consumer->unsubscribeInterface(m_prop);
      }
      const SDOPackage::NVList& m_prop;
    };

  private:
    std::vector<OutPortProvider*> m_providers;
    std::vector<InPortConsumer*> m_consumers;
    OutPortBase& m_outport;

    PublisherFactory m_pf;

    struct subscribe
    {
      subscribe(const ConnectorProfile& prof)
	: m_prof(&prof), _consumer(NULL) 
      {
      }

      subscribe(const subscribe& subs)
	: m_prof(subs.m_prof),
	  _consumer(subs._consumer)
      {
      }

      subscribe& operator=(const subscribe& subs)
      {
	if (this == &subs) return *this;
	m_prof = subs.m_prof;
	_consumer = subs._consumer;
	return *this;
      }

      void operator()(InPortConsumer* cons)
      {
	if (cons->subscribeInterface(m_prof->properties))
	  {
	    _consumer = cons;
	  }
      }
      const ConnectorProfile* m_prof;
      InPortConsumer* _consumer;
    };


  };
}; // namespace RTC

#endif // DataOutPort_h
