// -*- C++ -*-
/*!
 * @file Properties.h
 * @brief Property list class (derived from Java Properties)
 * @date $Date: 2007/04/24 01:24:32 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: Properties.h,v 1.7 2007/04/24 01:24:32 n-ando Exp $
 *
 */

/*
 * $Log: Properties.h,v $
 * Revision 1.7  2007/04/24 01:24:32  n-ando
 * createNode() was added.
 *
 * Revision 1.6  2007/04/23 04:56:43  n-ando
 * Some fixes about const.
 *
 * Revision 1.5  2007/04/13 18:14:07  n-ando
 * Now Properties class is implemented as hierarchical data structure
 * instead of std::map based data structure.
 *
 * Revision 1.4  2007/01/06 17:59:23  n-ando
 * getProperty() is now const function.
 *
 * Revision 1.3  2006/10/17 10:10:21  n-ando
 * Some escape/unescape related bugs are fixed.
 * The following some functions were moved to StringUtil.h
 * - eraseHeadBlank(), eraseTailBlank(), replaceString()
 *
 * Revision 1.2  2006/10/13 03:50:00  n-ando
 * The bug that white space delimiter did not work properly was fixed.
 *
 * Revision 1.1  2006/09/11 18:09:57  n-ando
 * Java compatible properties class.
 *
 *
 */

#ifndef Properties_h
#define Properties_h


#include <string>
#include <vector>
#include <map>


namespace RTC
{
  /*!
   * @if jp
   *
   * @class Properties
   *
   * Properties 饹ϡѤΥץѥƥåȤɽProperties 򥹥ȥ꡼
   * ݴɤꡢȥ꡼फɤꤹ뤳ȤǤ롣
   * ץѥƥꥹȤγƥӤбͤʸȤʤäƤ롣
   *
   * ץѥƥꥹȤˤϡΡ֥ǥե͡פȤ̤ΥץѥƥꥹȤ
   * ȤǤ롣ΥץѥƥꥹȤǥץѥƥĤʤȡ
   * 2ܤΥץѥƥꥹȤ롣 
   *
   * ץѥƥμˤ getPropertyץѥƥΥåȤˤ setProperty 
   * ä᥽åɤѤ뤳Ȥ侩롣
   *
   * ץѥƥ򥹥ȥ꡼¸Ȥޤϥȥ꡼फɤȤˡ
   * ISO 8859-1 ʸ󥳡ǥ󥰤Ѥ롣Υ󥳡ǥ󥰤ľɽ
   * ǤʤʸϡȤǤʤ
   *
   * Υ饹ϡJava  Properties 饹 (java.util.Properties) ȤۤƱͤ
   * ᥽åɤġޤϤե Java  Properties 饹
   * ϤΤȸߴ뤬Unicode ޤΤϰȤǤʤ
   *
   * @else
   *
   * @class Properties
   *
   * The Properties class represents a persistent set of properties. The
   * Properties can be saved to a stream or loaded from a stream. Each key and
   * its corresponding value in the property list is a string. 
   *
   * A property list can contain another property list as its "defaults"; this
   * second property list is searched if the property key is not found in the
   * original property list. 
   *
   * Because Properties inherits from Hashtable, the put and putAll methods can
   * be applied to a Properties object. Their use is strongly discouraged as they
   * allow the caller to insert entries whose keys or values are not Strings. The
   * setProperty method should be used instead. If the store or save method is
   * called on a "compromised" Properties object that contains a non-String key
   * or value, the call will fail. 
   *
   * The load and store methods load and store properties in a simple
   * line-oriented format specified below. This format uses the ISO 8859-1
   * character encoding. Characters that cannot be directly represented in this
   * encoding can be written using Unicode escapes ; only a single 'u' character
   * is allowed in an escape sequence. The native2ascii tool can be used to
   * convert property files to and from other character encodings. 
   *
   * This class has almost same methods of Java's Properties class. Input and 
   * Output stream of this properties are compatible each other except Unicode
   * encoded property file.
   *
   * @endif
   */
  class Properties
  {
  public:
    /*!
     * @if jp
     *
     * @brief 󥹥ȥ饯
     *
     * key  value ΤߤͿ Property Υ롼ȥΡɤ롣
     * ͤƥǥեͤȤꤵ롣
     * 
     * @else
     *
     * @brief Constructor
     *
     * Creates a root node of Property with root's key and value.
     *
     * @endif
     */
    Properties(const char* key = "", const char* value = "");
    
    
    /*!
     * @if jp
     *
     * @brief 󥹥ȥ饯
     *
     * std::string  std::map ǥեͤˤProperties롣
     * ͤƥǥեͤȤꤵ롣
     * 
     * @else
     *
     * @brief Constructor
     *
     * Creates an Properties with default value of std::string map.
     *
     * @endif
     */
    Properties(std::map<std::string, std::string>& defaults);
    
    
    /*!
     * @if jp
     *
     * @brief 󥹥ȥ饯
     *
     * ꤵ줿ǥեͤĶΥץѥƥꥹȤ롣
     * ͤƥǥեͤȤꤵ롣
     * ǥեͤ char* ˤͿ졢key  value ФˤʤäƤꡢ
     * ꥹȤνüοɽ num ʸ key Ϳʤ
     * ʤʤ
     * ʲ򼨤
     *
     * const char* defaults = {
     *     "key1", "value1",
     *     "key2", "value2",
     *     "key3", "value3",
     *     "key4", "value4",
     *     "key5", "value5",
     *     "" };
     * Properties p(defaults);
     * // ⤷
     * Properties p(defaults, 10);
     * 
     * @else
     *
     * @brief Constructor
     *
     * Creates an empty property list with the specified defaults.
     * The default values are given by array of char*, which should be pairs
     * of "key" and "value". The end of list is specified by argument "num",
     * which specifies number of array or null character of key.
     * The following is an example.
     *
     * const char* defaults = {
     *     "key1", "value1",
     *     "key2", "value2",
     *     "key3", "value3",
     *     "key4", "value4",
     *     "key5", "value5",
     *     "" };
     * Properties p(defaults);
     * // or
     * Properties p(defaults, 10);
     *
     * @endif
     */
    Properties(const char* defaults[], long num = LONG_MAX);
    
    
    /*!
     * @if jp
     * @brief ԡ󥹥ȥ饯
     *
     * Ϳ줿 Properties Υͤӥǥեͤ
     * ƤΤޤޥԡ롣
     *
     * @else
     *
     * @brief Copy Constructor
     *
     * All of given Properties's keys, values and default values 
     * are copied to new Properties.
     * 
     * @endif
     */
    Properties(const Properties& prop);


    /*!
     * @if jp
     * @brief 黻
     *
     * ͤ Properties Υͤӥǥեͤƺ졢
     * ͤ Properties ΥͤӥǥեͤƤΤޤ
     * ԡ롣
     *
     * @else
     *
     * @brief Assignment operator
     *
     *
     * @endif
     */
    Properties& operator=(const Properties& prop);
    

    /*!
     * @if jp
     *
     * @brief ǥȥ饯
     *
     * @else
     *
     * @brief Destructor
     *
     * @endif
     */
    virtual ~Properties();
    

    //============================================================
    // public functions
    //============================================================
    

    inline const char* getName() const          {return name.c_str();}
    inline const char* getVlue() const         {return value.c_str();}
    inline const char* getDefaultValue() const {return default_value.c_str();}
    inline const std::vector<Properties*>& getLeaf() const {return leaf;}
    inline const Properties* getRoot() const    {return root;}

    /*!
     * @if jp
     *
     * @brief ꤵ줿ĥץѥƥ򡢥ץѥƥꥹȤõ
     *
     * ꤵ줿ĥץѥƥ򡢥ץѥƥꥹȤõ
     * ΥץѥƥꥹȤˤʤȡǥեȤΥץѥƥꥹȡ
     * ˤΥǥե֤ͤĴ٤롣
     * ΥץѥƥĤʤϡnull ֤롣 
     *
     * @param key ץѥƥ
     * @return ꤵ줿ͤĤΥץѥƥꥹȤ
     *
     * @else
     *
     * @brief Searches for the property with the specified key in this property
     *
     * Searches for the property with the specified key in this property list.
     * If the key is not found in this property list, the default property list,
     * and its defaults, recursively, are then checked. The method returns null
     * if the property is not found. 
     *
     * @param key the property key.
     * @return the value in this property list with the specified key value.
     *
     * @endif
     */
    const std::string& getProperty(const std::string& key) const;
    

    /*!
     * @if jp
     *
     * @brief ꤵ줿ĥץѥƥ򡢥ץѥƥꥹȤõ 
     *
     * ꤵ줿ĥץѥƥ򡢥ץѥƥꥹȤõ
     * ΥץѥƥꥹȤˤʤȡǥեȤΥץѥƥꥹȡ
     * ˤΥǥե֤ͤĴ٤롣
     * ΥץѥƥĤʤϡǥեͤΰ֤롣 
     *
     * @param key ץѥƥ
     * @param defaultValue ǥե
     * @return ꤵ줿ͤĤΥץѥƥꥹȤ
     *
     * @else
     *
     * @brief Searches for the property with the specified key in this property
     *
     * Searches for the property with the specified key in this property list.
     * If the key is not found in this property list, the default property list,
     * and its defaults, recursively, are then checked. The method returns the
     * default value argument if the property is not found.
     *
     * @param key the property key
     * @param defaultValue a default value. 
     * @return the value in this property list with the specified key value.
     *
     * @endif
     */
    const std::string& getProperty(const std::string& key,
			     const std::string& def) const;

    /*!
     * @if jp
     *
     * @brief ꤵ줿ĥץѥƥ򡢥ץѥƥꥹȤõ 
     *
     * ꤵ줿ĥץѥƥ֤
     * ΥץѥƥꥹȤˤʤХǥե֤ͤ
     * ˸ĤʤСʸ֤
     *
     * @param key ץѥƥ
     * @return ꤵ줿ͤĤΥץѥƥꥹȤ
     *
     * @else
     *
     * @brief Searches for the property with the specified key in this property
     *
     * Searches for the property with the specified key in this property list.
     * If the key is not found in this property list, the default property list,
     * and its defaults, recursively, are then checked. The method returns the
     * default value argument if the property is not found.
     *
     * @param key the property key
     * @param defaultValue a default value. 
     * @return the value in this property list with the specified key value.
     *
     * @endif
     */
    const std::string& operator[](const std::string& key) const;


    /*!
     * @if jp
     *
     * @brief ꤵ줿ĥץѥƥ򡢥ץѥƥꥹȤõ 
     *
     * ꤵ줿ĥץѥƥ֤
     * ΥͤץѥƥꥹȤˤʤХǥե֤ͤ
     * ˸ĤʤСʸ֤
     * ͤˤʤˡƱͤǤʤȤͿ줿
     * бץѥƥ˱ͤ
     *
     * @param key ץѥƥ
     * @return ꤵ줿ͤĤΥץѥƥꥹȤ
     *
     * @else
     *
     * @brief Searches for the property with the specified key in this property
     *
     * Searches for the property with the specified key in this property list.
     * If the key is not found in this property list, the default property list,
     * and its defaults, recursively, are then checked. The method returns the
     * default value argument if the property is not found.
     *
     * @param key the property key
     * @param defaultValue a default value. 
     * @return the value in this property list with the specified key value.
     *
     * @endif
     */
    std::string& operator[](const std::string& key);

    
    /*!
     * @if jp
     * @brief ꤵ줿Фƥǥեͤꤹ
     * @else
     * @brief Set value as the default value to specified key's property
     * @endif
     */
    const std::string& getDefault(const std::string& key) const;

    
    /*!
     * @if jp
     *
     * @brief Properties  value  key ˤĤϿ
     *
     * Properties  value  key ˤĤϿ롣
     * Ǥ key ФͤäƤ硢ͤ˸Ť֤ͤ
     *
     * @param key ץѥƥꥹȤ֤륭
     * @param value key б 
     * @return ץѥƥꥹȤλꤵ줿͡줬ʤ null
     *
     * @else
     *
     * @brief Sets a value associated with key in the property list
     *
     * This method sets the "value" associated with "key" in the property list.
     * If the property list has a value of "key", old value is returned.
     *
     * @param key the key to be placed into this property list.
     * @param value the value corresponding to key. 
     * @return the previous value of the specified key in this property list,
     *         or null if it did not have one.
     *
     *@endif
     */
    std::string setProperty(const std::string& key, const std::string& value);


    /*!
     * @if jp
     * @brief Properties ˥ǥե value  key ˤĤϿ
     * @else
     * @brief Sets a default value associated with key in the property list
     * @endif
     */
    std::string setDefault(const std::string& key, const std::string& value);
    

    /*!
     * @if jp
     * @brief Properties ˥ǥեͤޤȤϿ
     * @else
     * @brief Sets a default value associated with key in the property list
     * @endif
     */
    void setDefaults(const char* defaults[], long num = LONG_MAX);

    
    //============================================================
    // load and save functions
    //============================================================
    /*!
     * @if jp
     *
     * @brief ꤵ줿ϥȥ꡼ˡץѥƥꥹȤϤ
     *
     * ꤵ줿ϥȥ꡼ˡץѥƥꥹȤϤ롣
     * Υ᥽åɤϼ˥ǥХåѤ롣
     *
     * @param out ϥȥ꡼
     *
     * @else
     *
     * @brief Prints this property list out to the specified output stream
     *
     * Prints this property list out to the specified output stream.
     * This method is useful for debugging.
     *
     * @param out an output stream.
     *
     * @endif
     */
    void list(std::ostream& out);
    
    
    /*!
     * @if jp
     *
     * @brief ϥȥ꡼फ饭ǤФˤʤäץѥƥꥹȤɤ߹
     *
     * ϥȥ꡼फ饭ǤФˤʤäץѥƥꥹȤɤ߹ࡣ
     * ȥ꡼ϡISO 8859-1 ʸ󥳡ǥ󥰤ѤƤȤߤʤ롣
     * ƥץѥƥϡϥȥ꡼˹ñ̤ϿƤΤȤߤʤ졢
     * ƹԤϹԶڤʸ (\n\rޤ \r\n) ǽ롣
     * ϥȥ꡼फɤ߹Ԥϡϥȥ꡼ǥեν
     * ãޤǽ롣
     *
     * ʸιԡޤϺǽʸ ASCII ʸ # ޤ ! Ǥ
     * Ԥ̵뤵롣Ĥޤꡢ# ޤ ! ϥȹԤ򼨤
     *
     * Ԥޤϥȹ԰ʳΤ٤ƤιԤϡơ֥ɲäץѥƥ
     * 򵭽Ҥ롣Ԥν꤬ \ ξϡιԤз³ԤȤ
     * ޤ (򻲾) ϡǽʸ顢ǽ ASCII ʸ
     * =:ޤ϶ʸľޤǤΡΤ٤Ƥʸ鹽롣
     *
     * ν򼨤ʸϡ \ դ뤳Ȥˤꥭ˴ޤ뤳Ȥ
     * Ǥ롣θζϤ٤ƥåפ롣
     * θκǽʸ = ޤ : ǤϡΥ
     * ̵뤵졢ΤȤζʸ⤹٤ƥåפ롣
     * ΤʳʸϤ٤ơϢʸΰȤʤ롣
     * ʸǤϡASCII ץ \t\n\r\\\"\'
     * \ (ߵȥڡ) \uxxxx ǧ졢ñȤʸѴ롣
     * ޤԤκǸʸ \ ǤϡιԤϸߤιԤη³Ȥ
     * 롣ξ硢\ ȹԶڤʸ˴졢³ԤƬ˶
     * Ф⤹٤˴졢ʸΰˤϤʤʤ 
     *
     * ȤС 4 ԤϤ줾쥭 Truth ȡϢ Beauty ɽ
     * 
     * Truth = Beauty <BR>
     *	Truth:Beauty <BR>
     * Truth			:Beauty <BR>
     *
     * ޤ 3 Ԥ 1 ĤΥץѥƥɽ 
     *
     * fruits				apple, banana, pear, \ <BR>
     *                                  cantaloupe, watermelon, \ <BR>
     *                                  kiwi, mango <BR>
     *  fruits ǡǤ˴Ϣդ롣 
     * "apple, banana, pear, cantaloupe, watermelon, kiwi, mango"
     * ǽŪʷ̤ǥޤΤȤɬڡɽ褦ˡ
     *  \ ˥ڡ롣Ԥν򼨤 \ ȡ³ԤƬˤ
     * ˴졢¾ʸִʤ 
     * ޤ 3 ܤǤϡ cheeses ǡϢǤʸ
     * Ǥ뤳Ȥɽ 
     *
     * cheeses <BR>
     * ϡcheeses ǡϢǤ϶ʸǤ뤳ȤꤷƤ롣 
     *
     * @param inStream ϥȥ꡼ 
     *
     * @else
     *
     * @brief Loads property list consists of key:value from input stream
     *
     * Reads a property list (key and element pairs) from the input stream.
     * The stream is assumed to be using the ISO 8859-1 character encoding; that
     * is each byte is one Latin1 character. Characters not in Latin1, and
     * certain special characters, can be represented in keys and elements using
     * escape sequences similar to those used for character and string literals
     * The differences from the character escape sequences used for characters
     * and strings are: 
     * - Octal escapes are not recognized. 
     * - The character sequence \b does not represent a backspace character. 
     * - The method does not treat a backslash character, \, before a non-valid
     *   escape character as an error; the backslash is silently dropped. For
     *   example, in a Java string the sequence "\z" would cause a compile time
     *   error. In contrast, this method silently drops the backslash. Therefore,
     *   this method treats the two character sequence "\b" as equivalent to the
     *   single character 'b'. 
     * - Escapes are not necessary for single and double quotes; however, by the
     *   rule above, single and double quote characters preceded by a backslash
     *   still yield single and double quote characters, respectively. 
     * An IllegalArgumentException is thrown if a malformed Unicode escape
     * appears in the input. 
     *
     * This method processes input in terms of lines. A natural line of input is
     * terminated either by a set of line terminator characters
     * (\n or \r or \r\n) or by the end of the file. A natural line may be either
     * a blank line, a comment line, or hold some part of a key-element pair.
     * The logical line holding all the data for a key-element pair may be spread
     * out across several adjacent natural lines by escaping the line terminator
     * sequence with a backslash character, \. Note that a comment line cannot be
     * extended in this manner; every natural line that is a comment must have
     * its own comment indicator, as described below. If a logical line is
     * continued over several natural lines, the continuation lines receive
     * further processing, also described below. Lines are read from the input
     * stream until end of file is reached. 
     *
     * A natural line that contains only white space characters is considered
     * blank and is ignored. A comment line has an ASCII '#' or '!' as its first
     * non-white space character; comment lines are also ignored and do not
     * encode key-element information. In addition to line terminators, this
     * method considers the characters space (' ', '\u0020'), tab 
     * ('\t', '\u0009'), and form feed ('\f', '\u000C') to be white space. 
     *
     * If a logical line is spread across several natural lines, the backslash
     * escaping the line terminator sequence, the line terminator sequence, and
     * any white space at the start the following line have no affect on the key
     * or element values. The remainder of the discussion of key and element
     * parsing will assume all the characters constituting the key and element
     * appear on a single natural line after line continuation characters have
     * been removed. Note that it is not sufficient to only examine the character
     * preceding a line terminator sequence to see if the line terminator is
     * escaped; there must be an odd number of contiguous backslashes for the
     * line terminator to be escaped. Since the input is processed from left to
     * right, a non-zero even number of 2n contiguous backslashes before a line
     * terminator (or elsewhere) encodes n backslashes after escape processing. 
     *
     * The key contains all of the characters in the line starting with the first
     * non-white space character and up to, but not including, the first
     * unescaped '=', ':', or white space character other than a line terminator.
     * All of these key termination characters may be included in the key by
     * escaping them with a preceding backslash character; for example,
     *
     * \:\=
     *
     * would be the two-character key ":=". Line terminator characters can be
     * included using \r and \n escape sequences. Any white space after the key
     * is skipped; if the first non-white space character after the key is '=' or
     * ':', then it is ignored and any white space characters after it are also
     * skipped. All remaining characters on the line become part of the
     * associated element string; if there are no remaining characters, the
     * element is the empty string "". Once the raw character sequences
     * constituting the key and element are identified, escape processing is
     * performed as described above. 
     *
     * As an example, each of the following three lines specifies the key "Truth"
     * and the associated element value "Beauty": 
     *
     * Truth = Beauty <BR>
     *        Truth:Beauty <BR>
     * Truth                  :Beauty <BR>
     *  As another example, the following three lines specify a single property: 
     *
     * fruits                           apple, banana, pear, \ <BR>
     *                                  cantaloupe, watermelon, \ <BR>
     *                                  kiwi, mango <BR>
     * The key is "fruits" and the associated element is: 
     * "apple, banana, pear, cantaloupe, watermelon, kiwi, mango"Note that a
     * space appears before each \ so that a space will appear after each comma
     * in the final result; the \, line terminator, and leading white space on
     * the continuation line are merely discarded and are not replaced by one or
     * more other characters. 
     * As a third example, the line: 
     *
     * cheeses <BR>
     * specifies that the key is "cheeses" and the associated element is the
     * empty string "".
     *
     * @param inStream the input stream.
     *
     * @endif
     */
    void load(std::istream& inStream);
    
    
    /*!
     * @if jp
     *
     * @brief ץѥƥꥹȤꤵ줿ȥ꡼¸
     *
     * 侩Ƥޤ󡣥ץѥƥꥹȤ¸ˡȤƤϡ
     * store(ostream out, string header) ᥽åɤλѤ侩롣
     * Υ᥽åɤ Java Properties ȤθߴΤƤ롣
     *
     * @param out ϥȥ꡼
     * @param header ץѥƥꥹȤε 
     *
     * @else
     *
     * @brief Save the properties list to the stream
     *
     * Deprecated. 
     *
     * @param out The output stream
     * @param header A description of the property list
     *
     * @endif
     */
    void save(std::ostream& out, const std::string& header);
    
    
    /*!
     * @if jp
     *
     * @brief ץѥƥꥹȤϥȥ꡼¸
     *
     * Properties ơ֥Υץѥƥꥹ (ǤΥڥ) load
     * ᥽åɤȤä Properties ơ֥˥ɤΤŬڤʥեޥåȤ
     * ϥȥ꡼˽񤭹ߤޤ 
     *
     * Properties ơ֥Υץѥƥꥹ (ǤΥڥ) load
     * ᥽åɤȤä Properties ơ֥˥ɤΤŬڤʥեޥåȤ
     * ϥȥ꡼˽񤭹ߤޤȥ꡼ϡISO 8859-1 ʸ
     * 󥳡ǥ󥰤Ѥƽ񤭹ޤޤ 
     * Properties ơ֥ (¸ߤ) Υǥեȥơ֥뤫
     * ץѥƥϡΥ᥽åɤˤäƤϽ񤭹ޤޤ 
     *
     * header  null ǤʤϡASCII ʸ #header ʸ
     * ӹԶڤʸǽ˽ϥȥ꡼˽񤭹ޤޤΤᡢ
     * header ϼ̥ȤȤƻȤȤǤޤ 
     *
     * ˡASCII ʸ #ߤ (Date  toString ᥽åɤˤä
     * ߻郎ΤƱ) Writer ˤäԶڤ
     * ʤ륳ȹԤ񤭹ޤޤ 
     *
     * ³ơProperties ơ֥Τ٤ƤΥȥ꤬ 1 ԤĽ񤭽Фޤ
     * ƥȥΥʸASCII ʸ =Ϣʸ󤬽񤭹ޤ
     * ʸγʸϡץ󥹤Ȥ褹ɬפ뤫
     * ɤǧޤASCII ʸ \֡ԡϤ줾 \\
     * \t\n \r Ȥƽ񤭹ޤޤ\u0020 꾮ʸ
     * \u007E 礭ʸϡб 16  xxxx Ȥä \uxxxx Ȥƽ
     * ޤޤ߶ʸǤ񤭶ʸǤʤԶʸϡ
     *  \ դƽ񤭹ޤޤͤʸ #!= : ϡ
     * ɬɤ褦ˡ˥åդƽ񤭹ޤޤ 
     *
     * ȥ꤬񤭹ޤ줿Ȥǡϥȥ꡼बեå夵ޤ
     * ϥȥ꡼ϤΥ᥽åɤȤⳫޤޤǤ 
     *
     * @else
     *
     * @brief Stores property list to the output stream
     *
     * Writes this property list (key and element pairs) in this Properties table
     * to the output stream in a format suitable for loading into a Properties
     * table using the load method. The stream is written using the ISO 8859-1
     * character encoding. 
     *
     * Properties from the defaults table of this Properties table (if any) are
     * not written out by this method. 
     *
     * If the comments argument is not null, then an ASCII # character, the
     * comments string, and a line separator are first written to the output
     * stream. Thus, the comments can serve as an identifying comment. 
     *
     * Next, a comment line is always written, consisting of an ASCII #
     * character, the current date and time (as if produced by the toString
     * method of Date for the current time), and a line separator as generated
     * by the Writer. 
     *
     * Then every entry in this Properties table is written out, one per line.
     * For each entry the key string is written, then an ASCII =, then the
     * associated element string. Each character of the key and element strings
     * is examined to see whether it should be rendered as an escape sequence.
     * The ASCII characters \, tab, form feed, newline, and carriage return are
     * written as \\, \t, \f \n, and \r, respectively. Characters less than
     * \u0020 and characters greater than \u007E are written as \uxxxx for the
     * appropriate hexadecimal value xxxx. For the key, all space characters are
     * written with a preceding \ character. For the element, leading space
     * characters, but not embedded or trailing space characters, are written
     * with a preceding \ character. The key and element characters #, !, =, and
     * : are written with a preceding backslash to ensure that they are properly
     * loaded. 
     *
     * After the entries have been written, the output stream is flushed. The
     * output stream remains open after this method returns. 
     *
     * @param out an output stream.
     * @param header a description of the property list.
     *
     * @endif
     */
    void store(std::ostream& out, const std::string& header);

    
    //============================================================
    // other util functions
    //============================================================
    /*!
     * @if jp
     *
     * @brief ץѥƥΥΥꥹȤ vector ֤
     *
     * ᥤץѥƥꥹȤƱ̾ΥĤʤϡǥեȤ
     * ץѥƥꥹȤˤ̤ΥޤࡢΥץѥƥꥹȤˤ뤹٤
     * ΥΥꥹȤ֤ޤ 
     *
     * @return ץѥƥꥹȤˤ뤹٤ƤΥΥꥹȡ
     *         ǥեȤΥץѥƥꥹȤˤ륭ޤ
     *
     * @else
     *
     * @brief Returns an vector of all the keys in this property
     *
     * Returns an enumeration of all the keys in this property list, including
     * distinct keys in the default property list if a key of the same name has
     * not already been found from the main properties list.
     *
     * @return an vector of all the keys in this property list, including the
     *         keys in the default property list.
     *
     * @endif
     */
    std::vector<std::string> propertyNames() const;


    /*!
     * @if jp
     * @brief ץѥƥο
     * @else
     * @brief Get number of Properties
     * @endif
     */
    int size() const;


    /*!
     * @if jp
     * @brief Ρɤ
     * @else
     * @brief Get node of Properties
     * @endif
     */
    Properties* getNode(const std::string& key) const;
    bool createNode(const char* key)
    {
      Properties* p(getNode(key));
      if (p != NULL) return false;
      (*this)[key] = "";
      return true;
    }

    /*!
     * @if jp
     * @brief ΡɤǤ
     * @else
     * @brief Get node of Properties
     * @endif
     */
    Properties* removeNode(const char* leaf_name);


    /*!
     * @if jp
     * @brief ҥΡɤkey뤫ɤ
     * @else
     * @brief If key exists in the children
     * @endif
     */
    Properties* hasKey(const char* key) const;


    /*!
     * @if jp
     * @brief ҥΡɤƺ
     * @else
     * @brief If key exists in the children
     * @endif
     */
    void clear();


    /*!
     * @if jp
     * @brief Propertyޡ
     * @else
     * @brief Merge properties
     * @endif
     */
    Properties& operator<<(const Properties& prop);
    

  protected:
    static void splitKeyValue(const std::string& str, std::string& key,
			      std::string& value);

    static bool split(const std::string& str, const char delim,
		      std::vector<std::string>& value);

    static Properties* _getNode(std::vector<std::string>& keys,
				std::vector<Properties*>::size_type index,
				const Properties* curr);

    static void _propertiyNames(std::vector<std::string>& names,
				std::string curr_name,
				const Properties* curr);

    static void _store(std::ostream& out, std::string curr_name,
		       Properties* curr);

    static std::ostream& _dump(std::ostream& out, const Properties& curr,
			       int index);

    static std::string indent(int index);
    
  private:
    std::string name;
    std::string value;
    std::string default_value;
    Properties* root;
    std::vector<Properties*> leaf;
    const std::string m_empty;
    friend std::ostream& operator<<(std::ostream& lhs, const Properties& rhs);

  };   // class Properties
};     // namespace RTC  
#endif // Properties_h
  
